#!/usr/bin/python3
#
# Copyright 2022-2023 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.
#
import asyncio
import json
import logging
import sys
import time
from typing import Any, Dict, List

import click
from fastapi import HTTPException

try:
    from idbroker.initial_sync import CONFIG_PATH, get_ous, read_id_broker_config, setup, sync_complete
except ImportError:  # pragma: no cover
    sys.path.append(
        "/var/lib/univention-appcenter/apps/ucsschool-id-connector/conf/plugins/packages"
    )  # pragma: no cover
    from idbroker.initial_sync import (
        setup,
        sync_complete,
        CONFIG_PATH,
        get_ous,
        read_id_broker_config,
    )  # pragma: no cover

from ucsschool_id_connector.http_api import query_service
from ucsschool_id_connector.models import SchoolAuthorityConfigurationPatchDocument
from ucsschool_id_connector.school_scheduler import SchoolScheduler
from ucsschool_id_connector.utils import ConsoleAndFileLogging

QUEUE_CHECK_TIMEOUT = 500
QUEUE_CHECK_INTERVAL = 3


class ManageSchoolsException(Exception):
    ...


class ManageSchoolsOnIDBroker:
    def __init__(
        self,
        school_authority: str,
        logger: logging.Logger,
        force: bool,
        num_tasks_for_initial_sync: int = 0,
    ):
        self.num_tasks = num_tasks_for_initial_sync
        self.school_authority = school_authority
        self.logger = logger
        self.school_scheduler = SchoolScheduler()
        self.force = force
        setup_logging(
            loggers=[
                self.logger,
                self.school_scheduler.logger,
                self.school_scheduler.user_scheduler.logger,
                self.school_scheduler.group_scheduler.logger,
            ]
        )
        self.old_school_authority_config = None

    def safe_current_config(self, school_authority: str):
        """Save old state for possible rollbacks."""
        self.old_school_authority_config = read_id_broker_config(
            config_path=CONFIG_PATH, school_authority_name=school_authority
        )

    def restore_old_config(self, school_authority: str):
        config = self.old_school_authority_config.dict_secrets_as_str()
        with open(f"{CONFIG_PATH}/{school_authority}.json", "w") as fout:
            json.dump(fp=fout, obj=config)

    async def add_school_to_config(self, school: str):
        id_broker_config = self.old_school_authority_config.dict()
        schools = set(
            school for school in id_broker_config["plugin_configs"]["id_broker"].get("schools", [])
        )
        _lower_schools = [school.lower() for school in schools]
        if school.lower() in _lower_schools and self.force is False:
            raise ManageSchoolsException(
                f"School {school} was already added (use --force if you want to resync the school)"
            )
        schools.add(school)
        patch_data = {"plugin_configs": {"id_broker": {"schools": list(schools)}}}
        await self.patch_school_authority_config(
            data=patch_data,
        )
        self.safe_current_config(school_authority=self.school_authority)

    async def patch_school_authority_config(self, data: Dict[str, Any]):
        patch_data = SchoolAuthorityConfigurationPatchDocument(**data)
        try:
            res = await query_service(
                cmd="patch_school_authority", name=self.school_authority, school_authority=patch_data
            )
            if res.get("errors"):
                raise ManageSchoolsException(f"Error while trying to patch the school authority: {res}")
        except HTTPException as e:
            raise ManageSchoolsException(f"Error while trying to patch the school authority: {e!r}")

    async def set_initial_sync_mode(self, value: bool):
        patch_data = {"plugin_configs": {"id_broker": {"initial_import_mode": value}}}
        await self.patch_school_authority_config(
            data=patch_data,
        )

    async def remove_school_from_config(self, school: str):
        id_broker_config = self.old_school_authority_config.dict()
        schools = set(id_broker_config["plugin_configs"]["id_broker"].get("schools", []))
        if school in schools:
            schools.remove(school)
        patch_data = {"plugin_configs": {"id_broker": {"schools": list(schools)}}}
        await self.patch_school_authority_config(data=patch_data)
        self.safe_current_config(school_authority=self.school_authority)

    async def wait_for_empty_queue(self):
        start = time.time()
        while time.time() < start + QUEUE_CHECK_TIMEOUT:
            res = await query_service(cmd="get_queue", name=self.school_authority)
            if res.get("errors"):
                raise ManageSchoolsException(
                    f"Error while trying to get the length of queue for school authority: {self.school_authority}"
                )
            queue_length = res["result"]["length"]
            if queue_length > 0:
                self.logger.info(f"{queue_length} Remaining objects.")
                await asyncio.sleep(QUEUE_CHECK_INTERVAL)
            else:
                return
        raise ManageSchoolsException(
            f"School still exists on ID Broker after {QUEUE_CHECK_TIMEOUT}s."
            " This can happen when deleting large schools."
        )

    async def add_schools(
        self,
        schools: List[str],
    ):
        """Add the school to the config and schedule all existing users & groups.
        If the school is not existing, the second step won't do anything."""
        for school in schools:
            try:
                schools = await get_ous(ou=school)
                if len(schools) == 0:
                    raise ManageSchoolsException(f"School {school} does not exist on this system.")

                self.logger.info(f"Connect school {school} to the ID Broker.")
                self.logger.info(f"Adding {school} to the schools which are synced to the ID Broker.")
                self.safe_current_config(school_authority=self.school_authority)
                await self.add_school_to_config(school=school)
                self.logger.info(
                    f"Sync users of school {school} without groups using the ID Connector. "
                    "This might take a while."
                )
                await self.set_initial_sync_mode(value=True)
                await self.school_scheduler.queue_school(school=school, num_tasks=self.num_tasks)
                self.logger.info(
                    f"Waiting until all users of {school} are synced. This will take a while."
                )
                await self.wait_for_empty_queue()
                setup(config_path=CONFIG_PATH)
                self.logger.info(f"Sync groups of school {school} (without members).")
                await sync_complete(
                    school=school,
                    dry_run=False,
                    max_workers=self.num_tasks,
                    sync_members=False,
                )
                self.logger.info(f"Update group memberships of school {school}.")
                await sync_complete(
                    school=school,
                    dry_run=False,
                    max_workers=self.num_tasks,
                    sync_members=True,
                )
                await self.set_initial_sync_mode(value=False)
            except (ManageSchoolsException, KeyboardInterrupt) as e:
                self.logger.error(f"An error was raised while trying to add the schools: {e}")
                self.logger.warning("Restore old config.")
                self.restore_old_config(school_authority=self.school_authority)
                sys.exit(1)
            except Exception:
                self.restore_old_config(school_authority=self.school_authority)
                raise

    async def remove_schools(self, schools: List[str]):
        self.safe_current_config(school_authority=self.school_authority)
        for school in schools:
            try:
                self.logger.info(f"Removing school {school}.")
                await self.remove_school_from_config(school=school)
                # todo use provisioning api to delete school #100
                # todo handle according to self.force
                # todo option to only remove from config
            except (ManageSchoolsException, KeyboardInterrupt) as e:
                self.logger.error(f"An error was raised while trying to remove the schools: {e}")
                self.logger.warning("Restore old config.")
                self.restore_old_config(school_authority=self.school_authority)
                sys.exit(1)
            except Exception:
                self.restore_old_config(school_authority=self.school_authority)
                raise
        self.logger.info("Done removing schools.")

    async def configure_schools(self, schools: List[str] = None):
        self.safe_current_config(school_authority=self.school_authority)
        try:
            self.logger.info(
                f"Adding {len(schools)} schools to the schools which are synced to the ID Broker."
            )
            for school in schools:
                await self.add_school_to_config(school=school)
            self.logger.info(f"{[school for school in schools]}")
            self.logger.info(f"Configured {len(schools)} schools.")
        except (ManageSchoolsException, KeyboardInterrupt) as e:
            self.logger.error(f"An error was raised while trying to add the schools: {e}")
            self.logger.warning("Restore old config.")
            self.restore_old_config(school_authority=self.school_authority)
            sys.exit(1)
        except Exception:
            self.restore_old_config(school_authority=self.school_authority)
            raise


@click.group()
def manage_schools_cli():
    pass  # pragma: no cover


@manage_schools_cli.command(
    name="remove_schools",
    help="Remove schools from the set of schools which are synchronized to the ID Broker "
    "and remove them from the ID Broker system.",
)
@click.option("-a", "--school_authority", required=True, type=str)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Add this option to also handle already removed schools.",
)
@click.option(
    "-f",
    "--all_schools",
    is_flag=True,
    default=False,
    help="Add this option to configure all schools.",
)
@click.argument("schools", required=False, nargs=-1)
def remove_schools(schools: List[str], school_authority: str, force: bool, all_schools: bool):
    logger = ConsoleAndFileLogging.get_logger(__name__)
    manager = ManageSchoolsOnIDBroker(school_authority=school_authority, logger=logger, force=force)
    if all_schools:
        if schools:
            logger.error("Passing schools with --all_schools option is not permitted.")
            sys.exit(1)
        schools = asyncio.run(get_ous("*"))
        schools = [school.name for school in schools]
    asyncio.run(
        manager.remove_schools(
            schools=schools,
        )
    )


def setup_logging(loggers: List[logging.Logger]):
    for logger in loggers:
        ConsoleAndFileLogging.add_console_handler(logger)


@manage_schools_cli.command(
    name="add_schools",
    help="Configure schools to the ID Broker and synchronize the respective school objects.",
)
@click.option(
    "-t",
    "--num_tasks",
    type=click.IntRange(1, 32),
    default=1,
    help="The number of parallel tasks used for running the initial sync.",
)
@click.option("-a", "--school_authority", required=True, type=str)
@click.option(
    "-i",
    "--initial_sync",
    type=bool,
    default=True,
    help="Synchronize objects of the given schools actively to the ID Broker "
    "and configure them to the ID Broker."
    "If set to false, existing schools are configured to the ID Broker,"
    " but the school objects are not actively synchronized. "
    "To apply this for all schools, omid the schools argument."
    "If the schools do not yet exist on the ID Broker system, "
    "they need to be triggered manually by using the schedule_school script.",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Add this option to also handle already synchronized schools.",
)
@click.option(
    "-f",
    "--all_schools",
    is_flag=True,
    default=False,
    help="Add this option to configure all schools.",
)
@click.argument("schools", nargs=-1)
def add_schools(
    num_tasks: int,
    schools: List[str],
    school_authority: str,
    initial_sync: bool,
    force: bool,
    all_schools: bool,
):
    logger = ConsoleAndFileLogging.get_logger(__name__)
    manager = ManageSchoolsOnIDBroker(
        school_authority=school_authority,
        logger=logger,
        num_tasks_for_initial_sync=num_tasks,
        force=force,
    )
    if all_schools:
        if schools:
            logger.error("Passing schools with --all_schools option is not permitted.")
            sys.exit(1)
        schools = asyncio.run(get_ous("*"))
        schools = [school.name for school in schools]
    if initial_sync:
        asyncio.run(
            manager.add_schools(
                schools=schools,
            )
        )
    else:
        asyncio.run(manager.configure_schools(schools=schools))


if __name__ == "__main__":
    manage_schools_cli()
