# -*- coding: utf-8 -*-

# Copyright 2022-2023 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

from unittest.mock import AsyncMock, Mock, call, mock_open, patch

import pytest
from click.testing import CliRunner

from idbroker.initial_sync import CONFIG_PATH
from idbroker.manage_schools_to_sync import (
    ManageSchoolsException,
    ManageSchoolsOnIDBroker,
    add_schools,
    remove_schools,
)
from ucsschool_id_connector.models import SchoolAuthorityConfigurationPatchDocument


@pytest.fixture(autouse=True)
def patch_env(monkeypatch):
    monkeypatch.setenv("DOCKER_HOST_NAME", "test.hostname")
    monkeypatch.setenv("ldap_base", "dc=ldap,dc=base")
    monkeypatch.setenv("ldap_server_name", "TEST-SERVER-NAME")
    monkeypatch.setenv("ldap_server_port", "789")


def school_auth_config():
    return {
        "name": "Traeger2",
        "active": True,
        "url": "https://provisioning1.broker.test/",
        "plugins": ["id_broker-users", "id_broker-groups"],
        "plugin_configs": {
            "id_broker": {
                "password": "psw",
                "schools": [],
                "username": "provisioning-Traeger2",
                "version": 1,
            }
        },
    }


@patch(
    "idbroker.manage_schools_to_sync.query_service",
    AsyncMock(return_value={"errors": "some-error"}),
)
@pytest.mark.asyncio
async def test_patch_school_authority_config_fails():
    manager = ManageSchoolsOnIDBroker(school_authority="test_authority", logger=Mock(), force=False)
    with pytest.raises(ManageSchoolsException):
        await manager.patch_school_authority_config(data={})


@patch(
    "idbroker.manage_schools_to_sync.query_service",
    AsyncMock(return_value={"result": school_auth_config()}),
)
@pytest.mark.asyncio
async def test_patch_school_authority_config():
    manager = ManageSchoolsOnIDBroker(school_authority="test_authority", logger=Mock(), force=False)
    await manager.patch_school_authority_config(data=school_auth_config())


@pytest.mark.asyncio
@pytest.mark.parametrize("value", (True, False))
async def test_set_initial_sync_mode(value):
    manager = ManageSchoolsOnIDBroker(school_authority="test_authority", logger=Mock(), force=False)
    manager.patch_school_authority_config = AsyncMock()
    await manager.set_initial_sync_mode(value=value)
    expected_data = {"plugin_configs": {"id_broker": {"initial_import_mode": value}}}
    manager.patch_school_authority_config.assert_called_with(data=expected_data)


def test__safe_current_config():
    manager = ManageSchoolsOnIDBroker(school_authority="test_authority", logger=Mock(), force=False)
    with patch("idbroker.manage_schools_to_sync.read_id_broker_config") as read_id_broker_config_mock:
        manager.safe_current_config(school_authority="test_authority")
        read_id_broker_config_mock.assert_called_with(
            config_path=CONFIG_PATH, school_authority_name="test_authority"
        )


def test__restore_old_config():
    config = school_auth_config()
    manager = ManageSchoolsOnIDBroker(school_authority="test_authority", logger=Mock(), force=False)
    config["plugin_configs"]["id_broker"]["schools"] = ["myschool"]
    manager.old_school_authority_config = SchoolAuthorityConfigurationPatchDocument(**config)
    with patch("builtins.open", mock_open()) as m, patch("json.dump"):
        manager.restore_old_config(school_authority="test_authority")
        m.assert_called_with(f"{CONFIG_PATH}/{'test_authority'}.json", "w")


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "force,already_existing",
    [(True, False), (True, True), (False, False), (False, True)],
)
async def test_add_school_to_config(force, already_existing):
    manager = ManageSchoolsOnIDBroker(school_authority="test_authority", logger=Mock(), force=force)
    config = school_auth_config()
    school = "myschool"
    if already_existing:
        config["plugin_configs"]["id_broker"]["schools"] = [school]
    school_authority_config = SchoolAuthorityConfigurationPatchDocument(**config)
    manager.old_school_authority_config = school_authority_config
    manager.patch_school_authority_config = AsyncMock()
    ManageSchoolsOnIDBroker.safe_current_config = Mock()
    if already_existing is False or force is True:
        await manager.add_school_to_config(
            school=school,
        )
        data = {"plugin_configs": {"id_broker": {"schools": [school]}}}
        manager.patch_school_authority_config.assert_called_with(data=data)
        ManageSchoolsOnIDBroker.safe_current_config.assert_called_with(school_authority="test_authority")
    else:
        with pytest.raises(ManageSchoolsException):
            await manager.add_school_to_config(
                school=school,
            )
        manager.patch_school_authority_config.assert_not_called()


class Scheduler:
    def __init__(self):
        self.school = None
        self.num_tasks = None
        self.logger = Mock()
        self.user_scheduler = Mock()
        self.group_scheduler = Mock()

    async def queue_school(self, school, num_tasks):
        self.school = school
        self.num_tasks = num_tasks


@pytest.mark.asyncio
async def test_add_school():
    manager = ManageSchoolsOnIDBroker(
        school_authority="test_authority", logger=Mock(), num_tasks_for_initial_sync=20, force=False
    )
    manager.set_initial_sync_mode = AsyncMock()
    ManageSchoolsOnIDBroker.safe_current_config = Mock()
    ManageSchoolsOnIDBroker.restore_old_config = Mock()
    manager.add_school_to_config = AsyncMock()
    manager.wait_for_empty_queue = AsyncMock()
    manager.school_scheduler = AsyncMock()
    school_name = "myschool"
    with patch("idbroker.manage_schools_to_sync.setup") as setup_mock, patch(
        "idbroker.manage_schools_to_sync.sync_complete"
    ) as sync_complete_mock, patch(
        "idbroker.manage_schools_to_sync.get_ous", return_value=get_fake_schools()
    ):
        await manager.add_schools(schools=[school_name])
    assert manager.num_tasks == 20
    manager.add_school_to_config.assert_called_with(
        **{
            "school": school_name,
        }
    )
    assert manager.set_initial_sync_mode.call_count == 2
    assert manager.set_initial_sync_mode.call_args_list == [
        call(
            **{
                "value": True,
            }
        ),
        call(
            **{
                "value": False,
            }
        ),
    ]
    setup_mock.assert_called_with(config_path=CONFIG_PATH)
    assert sync_complete_mock.call_count == 2
    assert sync_complete_mock.call_args_list == [
        call(
            **{
                "school": school_name,
                "dry_run": False,
                "max_workers": 20,
                "sync_members": False,
            }
        ),
        call(
            **{
                "school": school_name,
                "dry_run": False,
                "max_workers": 20,
                "sync_members": True,
            }
        ),
    ]
    manager.wait_for_empty_queue.assert_called()


@pytest.mark.asyncio
async def test_add_school_school_does_not_exist():
    manager = ManageSchoolsOnIDBroker(
        school_authority="test_authority", logger=Mock(), num_tasks_for_initial_sync=20, force=False
    )
    manager.set_initial_sync_mode = AsyncMock()
    ManageSchoolsOnIDBroker.safe_current_config = Mock()
    ManageSchoolsOnIDBroker.restore_old_config = Mock()
    manager.add_school_to_config = AsyncMock()
    manager.wait_for_empty_queue = AsyncMock()
    manager.school_scheduler = AsyncMock()
    school_name = "nonexistingschool"
    with patch("idbroker.manage_schools_to_sync.setup"), patch(
        "idbroker.manage_schools_to_sync.sync_complete"
    ), patch("idbroker.manage_schools_to_sync.get_ous", return_value=get_fake_schools()):
        await manager.add_schools(schools=[school_name])


class MySpecialException(Exception):
    ...


def test_add_schools_script():
    runner = CliRunner()
    with patch("idbroker.manage_schools_to_sync.setup_logging"), patch.object(
        ManageSchoolsOnIDBroker, "add_schools"
    ) as add_school_mock:
        result = runner.invoke(
            add_schools,
            [
                "--num_tasks",
                2,
                "--school_authority",
                "test-authority",
                "school1",
                "school2",
            ],
        )
        add_school_mock.assert_called_with(schools=("school1", "school2"))
    assert not result.exception


@pytest.mark.asyncio
async def test_remove_school_from_config():
    config = {
        "name": "Traeger2",
        "active": True,
        "url": "https://provisioning1.broker.test/",
        "plugins": ["id_broker-users", "id_broker-groups"],
        "plugin_configs": {
            "id_broker": {
                "password": "psw",
                "schools": ["myschool"],
                "username": "provisioning-Traeger2",
                "version": 1,
            }
        },
    }
    manager = ManageSchoolsOnIDBroker(school_authority="test_authority", logger=Mock(), force=False)
    manager.old_school_authority_config = SchoolAuthorityConfigurationPatchDocument(**config)
    manager.patch_school_authority_config = AsyncMock()
    manager.safe_current_config = Mock()
    await manager.remove_school_from_config(school="myschool")
    args = manager.patch_school_authority_config.call_args
    assert args[1]["data"] == {"plugin_configs": {"id_broker": {"schools": []}}}


@pytest.mark.asyncio
async def test_remove_school():
    manager = ManageSchoolsOnIDBroker(school_authority="test_authority", logger=Mock(), force=False)
    manager.remove_school_from_config = AsyncMock()
    ManageSchoolsOnIDBroker.safe_current_config = Mock()
    ManageSchoolsOnIDBroker.restore_old_config = Mock()
    await manager.remove_schools(
        schools=["myschool"],
    )
    manager.remove_school_from_config.assert_called_with(
        school="myschool",
    )


@pytest.mark.parametrize(
    "exception",
    (MySpecialException(), ManageSchoolsException(), KeyboardInterrupt()),
    ids=lambda x: x.__class__,
)
@pytest.mark.parametrize("all_schools", (True, False), ids=lambda x: f"all_schools-{x}")
def test_add_school_exception_during_execution(exception, all_schools):
    runner = CliRunner()
    fake_ous = [FakeOu(school) for school in get_fake_schools()]
    ManageSchoolsOnIDBroker.safe_current_config = Mock()
    ManageSchoolsOnIDBroker.restore_old_config = Mock()
    with patch("idbroker.manage_schools_to_sync.setup_logging"), patch.object(
        ManageSchoolsOnIDBroker, "add_school_to_config", side_effect=exception
    ), patch("idbroker.manage_schools_to_sync.get_ous", return_value=fake_ous):
        if all_schools:
            args = [
                "--num_tasks",
                2,
                "--school_authority",
                "test-authority",
                "--all_schools",
            ]
        else:
            args = [
                "--num_tasks",
                2,
                "--school_authority",
                "test-authority",
                "school1",
                "school2",
            ]
        result = runner.invoke(
            add_schools,
            args,
        )
        if isinstance(exception, MySpecialException):
            assert isinstance(result.exception, MySpecialException)
        else:
            assert isinstance(result.exception, SystemExit)
    ManageSchoolsOnIDBroker.safe_current_config.assert_called_with(school_authority="test-authority")
    ManageSchoolsOnIDBroker.restore_old_config.assert_called_with(school_authority="test-authority")


@pytest.mark.parametrize(
    "exception",
    (MySpecialException(), ManageSchoolsException(), KeyboardInterrupt()),
    ids=lambda x: x.__class__,
)
@pytest.mark.parametrize("all_schools", (True, False), ids=lambda x: f"all_schools-{x}")
def test_remove_school_exception_during_execution(exception, all_schools):
    runner = CliRunner()
    ManageSchoolsOnIDBroker.safe_current_config = Mock()
    ManageSchoolsOnIDBroker.restore_old_config = Mock()
    fake_ous = [FakeOu(school) for school in get_fake_schools()]
    with patch("idbroker.manage_schools_to_sync.setup_logging"), patch.object(
        ManageSchoolsOnIDBroker, "remove_school_from_config", side_effect=exception
    ), patch("idbroker.manage_schools_to_sync.get_ous", return_value=fake_ous):
        if all_schools:
            args = ["--school_authority", "test-authority", "--all_schools"]
        else:
            args = ["--school_authority", "test-authority", "school1", "school2"]
        result = runner.invoke(
            remove_schools,
            args,
        )
    if isinstance(exception, MySpecialException):
        assert isinstance(result.exception, MySpecialException)
    else:
        assert isinstance(result.exception, SystemExit)
    ManageSchoolsOnIDBroker.safe_current_config.assert_called_with(school_authority="test-authority")
    ManageSchoolsOnIDBroker.restore_old_config.assert_called_with(school_authority="test-authority")


class FakeOu:
    def __init__(self, name):
        self.name = name


@pytest.mark.parametrize(
    "exception",
    (MySpecialException(), ManageSchoolsException(), KeyboardInterrupt()),
    ids=lambda x: x.__class__,
)
@pytest.mark.parametrize("all_schools", (True, False), ids=lambda x: f"all_schools-{x}")
def test_configure_schools_exception_during_execution(exception, all_schools):
    runner = CliRunner()
    ManageSchoolsOnIDBroker.safe_current_config = Mock()
    ManageSchoolsOnIDBroker.restore_old_config = Mock()
    fake_ous = [FakeOu(school) for school in get_fake_schools()]
    with patch("idbroker.manage_schools_to_sync.setup_logging"), patch.object(
        ManageSchoolsOnIDBroker, "add_school_to_config", side_effect=exception
    ), patch("idbroker.manage_schools_to_sync.get_ous", return_value=fake_ous):
        if all_schools:
            args = ["--school_authority", "test-authority", "--initial_sync", "false", "--all_schools"]
        else:
            args = ["--school_authority", "test-authority", "--initial_sync", "false", "school1"]
        result = runner.invoke(
            add_schools,
            args,
        )
    if isinstance(exception, MySpecialException):
        assert isinstance(result.exception, MySpecialException)
    else:
        assert isinstance(result.exception, SystemExit)
    ManageSchoolsOnIDBroker.safe_current_config.assert_called_with(school_authority="test-authority")
    ManageSchoolsOnIDBroker.restore_old_config.assert_called_with(school_authority="test-authority")


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "successful,length",
    [(False, 1), (True, 0)],
    ids=["existing-after-timeout", "not-existing-after-timeout"],
)
async def test_wait_for_empty_queue(successful, length):
    manager = ManageSchoolsOnIDBroker(
        school_authority="test_authority", logger=Mock(), num_tasks_for_initial_sync=20, force=False
    )
    with patch(
        "idbroker.manage_schools_to_sync.query_service",
        AsyncMock(return_value={"result": {"length": length}}),
    ), patch("idbroker.manage_schools_to_sync.QUEUE_CHECK_TIMEOUT", 1,), patch(
        "idbroker.manage_schools_to_sync.QUEUE_CHECK_INTERVAL",
        0.1,
    ):
        if successful:
            await manager.wait_for_empty_queue()
        else:
            with pytest.raises(ManageSchoolsException, match="School still exists on ID Broker"):
                await manager.wait_for_empty_queue()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "successful,length",
    [(False, 1), (True, 0)],
    ids=["existing-after-timeout", "not-existing-after-timeout"],
)
async def test_wait_for_empty_queue_errors_during_query(successful, length):
    manager = ManageSchoolsOnIDBroker(
        school_authority="test_authority", logger=Mock(), num_tasks_for_initial_sync=20, force=False
    )
    with patch(
        "idbroker.manage_schools_to_sync.query_service",
        AsyncMock(return_value={"errors": ["some error"]}),
    ), patch("idbroker.manage_schools_to_sync.QUEUE_CHECK_TIMEOUT", 1,), patch(
        "idbroker.manage_schools_to_sync.QUEUE_CHECK_INTERVAL",
        0.1,
    ):
        with pytest.raises(
            ManageSchoolsException, match="Error while trying to get the length of queue"
        ):
            await manager.wait_for_empty_queue()


def test_remove_schools_script():
    runner = CliRunner()
    with patch.object(ManageSchoolsOnIDBroker, "remove_schools") as mock:
        result = runner.invoke(
            remove_schools,
            [
                "--school_authority",
                "test-authority",
                "school1",
                "school2",
            ],
        )
        mock.assert_called_with(schools=("school1", "school2"))
    assert not result.exception


def get_fake_schools():
    return ["school1", "school2", "school3"]


@pytest.mark.asyncio
async def test_configure_all_schools():
    manager = ManageSchoolsOnIDBroker(school_authority="test-auth", logger=Mock(), force=False)
    manager.add_school_to_config = AsyncMock()
    manager.safe_current_config = Mock()
    manager.restore_old_config = Mock()
    config = school_auth_config()
    config["plugin_configs"]["id_broker"]["schools"] = ["school1", "school2", "school3"]
    school_authority_config = SchoolAuthorityConfigurationPatchDocument(**config)
    with patch("idbroker.manage_schools_to_sync.get_ous", return_value=get_fake_schools()), patch(
        "idbroker.manage_schools_to_sync.read_id_broker_config", return_value=school_authority_config
    ):
        await manager.configure_schools(schools=["school1", "school2", "school3"])
    assert manager.add_school_to_config.call_count == 3
    assert manager.add_school_to_config.call_args_list == [
        call(school=f"school{i}") for i in range(1, 4)
    ]


@pytest.mark.asyncio
async def test_configure_schools_by_name():
    manager = ManageSchoolsOnIDBroker(school_authority="test-auth", logger=Mock(), force=False)
    manager.add_school_to_config = AsyncMock()
    manager.safe_current_config = Mock()
    manager.restore_old_config = Mock()
    manager.get_school_authority_config = AsyncMock(return_value=school_auth_config())

    await manager.configure_schools(schools=["school1", "school2"])
    assert manager.add_school_to_config.call_count == 2
    assert manager.add_school_to_config.call_args_list == [
        call(school=f"school{i}") for i in range(1, 3)
    ]


def test_all_schools_without_initial_sync_script():
    runner = CliRunner()
    with patch.object(ManageSchoolsOnIDBroker, "configure_schools") as mock:
        result = runner.invoke(
            add_schools,
            ["--school_authority", "test-authority", "--initial_sync", "false"],
        )
        mock.assert_called_once()
        assert not result.exception
